#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : 2019-03-01 - 14:43:51
#
# pylint: disable=too-many-instance-attributes
"""
-----------

The :py:class:`FDTD` class is the heart of `nsfds2`. It coordinates all the
calculations that have to be led to perform the Finite Differences Time Domain
(FDTD) simulation. The :py:class:`FDTD` works with :

    * a configuration file (parsed by :py:class:`nsfds2.init.config.CfgSetup`)
    * initial fields (created by :py:class:`nsfds2.init.fields.Fields`)
    * a mesh (created with fdgrid)


Example
-------

::

    from nsfds2.init import CfgSetup, Fields
    from nsfds2.fdtd import FDTD
    from fdgrid.mesh import Mesh

    # Read simulation parameter in config file ~/nsfds2/nsfds2.conf
    cfg = CfgSetup()

    # Define the mesh
    msh = Mesh((cfg.nx, cfg.nz), (cfg.dx, cfg.dz), origin=(cfg.ix0, cfg.iz0), obstacles=[])

    # Init acoustic fields
    fld = Fields(msh, cfg)

    # Create simulation
    fdtd = FDTD(msh, fld, cfg)
    fdtd.run()



-----------
"""

import os
import sys
import time
import numpy as np
from progressbar import ProgressBar, Bar, ReverseBar, ETA
from nsfds2.utils import headers, check, timing, misc
from nsfds2.lib.efluxes import EulerianFluxes
from nsfds2.lib.vfluxes import ViscousFluxes
from nsfds2.lib.sfilter import SelectiveFilter
from nsfds2.lib.scapture import ShockCapture


class FDTD:
    """

    The Finite Differences Time Domain Solver

    Parameters
    ----------
        msh : :py:class:`nsfds2.fdgrid.mesh.Mesh`
            Mesh
        fld : :py:class:`nsfds2.init.fields.Fields`
            Initial fields
        cfg : :py:class:`nsfds2.init.config.CfgSetup`
            Configuration

    """

    def __init__(self, msh, fld, cfg):

        self.msh = msh
        self.fld = fld
        self.cfg = cfg

        # Check some of the simulation parameters
        check.Check(self.cfg, self.msh)

        # Init libs
        self.efluxes = EulerianFluxes(self.msh, self.fld, self.cfg)
        if self.cfg.vsc:
            self.vfluxes = ViscousFluxes(self.msh, self.fld, self.cfg)
        if self.cfg.flt:
            self.sfilter = SelectiveFilter(self.msh, self.fld, self.cfg)
        if self.cfg.cpt:
            self.scapture = ShockCapture(self.msh, self.fld, self.cfg)

        # Init probes
        if cfg.probes:
            self.probes = np.zeros((len(cfg.probes), cfg.ns))

        # Prompt user for start
        if not self.cfg.quiet:
            headers.start(cfg)

        # Max pressure
        v = (self.fld.ru.max()/self.cfg.rho0)**2 + (self.fld.rv.max()/self.cfg.rho0)**2
        self.pmax = - 0.5*self.cfg.rho0*v \
                    *(self.cfg.gamma - 2)/(self.cfg.gamma - 1) + self.cfg.S0

        # Init timings and residuals
        self.res = 0
        self.tt = 0
        timed_methods = ['total', 'efluxes', 'vfluxes', 'sfilt',
                         'scapt', 'save', 'pressure', 'probe']
        self.bench = {i: [] for i in timed_methods}
        self.tloopi = time.perf_counter()

    def _pre_run(self):

        # Save file
        self.fld.init_save()

        if not self.cfg.quiet:
            print('-'*int(self.columns))
            print('# Start main loop ')

    def run(self):
        """ Run the main loop. """

        self._pre_run()

        if not self.cfg.quiet and not self.cfg.timings:
            widgets = [Bar('>'), ' ', ETA(), ' ', ReverseBar('<')]
            pbar = ProgressBar(widgets=widgets, maxval=self.cfg.nt,
                               term_width=self.columns).start()

        for self.cfg.it in range(self.cfg.nt+1):

            self.tt = time.perf_counter()

            # FDTD
            self.eulerian_fluxes()
            self.viscous_flux()

            if self.cfg.mesh == 'curvilinear':
                self.num2phys()

            self.selective_filter()
            self.shock_capture()

            if self.cfg.mesh == 'curvilinear':
                self.phys2num()

            self.update_pressure()

            # Break when computation diverges
            self.check_results()

            # Display log
            self.log()

            # Update probes
            if self.cfg.probes:
                self.update_probes()

            # Progress bar
            if not self.cfg.timings and not self.cfg.quiet:
                pbar.update(self.cfg.it)


        self.fld.sfile.close()

        if not self.cfg.quiet and not self.cfg.timings:
            pbar.finish()

        if not self.cfg.quiet:
            print('-'*int(self.columns))
            msg = '# Simulation completed in {}.\n'
            msg += '# End at t = {:.4f} sec.'
            print(msg.format(misc.secs_to_dhms(time.perf_counter() - self.tloopi),
                             self.cfg.dt*self.cfg.it))
            print('-'*int(self.columns))

    @timing.proceed('efluxes')
    def eulerian_fluxes(self):
        """ Compute Eulerian fluxes. """

        self.efluxes.rk4()

    @timing.proceed('vfluxes')
    def viscous_flux(self):
        """ Viscous flux """
        if self.cfg.vsc:
            self.vfluxes.integrate()
            self.efluxes.cout()

    @timing.proceed('sfilt')
    def selective_filter(self):
        """ Selective filter """
        if self.cfg.flt:
            self.sfilter.apply()

    @timing.proceed('scapt')
    def shock_capture(self):
        """ Shock Capture """
        if self.cfg.cpt:
            self.scapture.apply()

    @timing.proceed('save')
    def save(self):
        """ Save data """

        if self.cfg.save:
            self.fld.sfile.create_dataset('rho_it' + str(self.cfg.it),
                                          data=self.fld.r, compression=self.cfg.comp)
            self.fld.sfile.create_dataset('rhou_it' + str(self.cfg.it),
                                          data=self.fld.ru, compression=self.cfg.comp)
            self.fld.sfile.create_dataset('rhov_it' + str(self.cfg.it),
                                          data=self.fld.rv, compression=self.cfg.comp)
            self.fld.sfile.create_dataset('rhoe_it' + str(self.cfg.it),
                                          data=self.fld.re, compression=self.cfg.comp)
        if self.cfg.probes:
            self.fld.sfile['probes_value'][:, self.cfg.it-self.cfg.ns:self.cfg.it] = self.probes

    @timing.proceed('pressure')
    def update_pressure(self):
        """ Update pressure field """

        self.fld.fdtools.p(self.fld.p, self.fld.r, self.fld.ru,
                           self.fld.rv, self.fld.re)

    @timing.proceed('probe')
    def update_probes(self):
        """ Update probes. """

        for n, c in enumerate(self.cfg.probes):
            self.probes[n, self.cfg.it%self.cfg.ns] = self.fld.p[c[0], c[1]]

    def check_results(self):
        """ Check if computation diverges. """

        if self.cfg.mesh == 'curvilinear':
            self.res = self.fld.fdtools.residual(self.fld.p*self.msh.J)
        else:
            self.res = self.fld.fdtools.residual(self.fld.p)

#        if (abs(self.res) > 100*self.pmax) or np.any(np.isnan(self.fld.p)):
        if np.any(np.isnan(self.fld.p)):

            print('Stop simulation at iteration ', self.cfg.it)
            if np.any(np.isnan(self.fld.p)):
                print('Nan : {}'.format(np.argwhere(np.isnan(self.fld.p))))

            self.fld.sfile.close()
            sys.exit(1)

    def phys2num(self):
        """ Convert curvilinear coordinates : from physical to numeric. """

        self.fld.r = self.fld.r/self.msh.J
        self.fld.ru = self.fld.ru/self.msh.J
        self.fld.rv = self.fld.rv/self.msh.J
        self.fld.re = self.fld.re/self.msh.J

    def num2phys(self):
        """ Convert curvilinear coordinates : from numeric to physical. """

        self.fld.r = self.fld.r*self.msh.J
        self.fld.ru = self.fld.ru*self.msh.J
        self.fld.rv = self.fld.rv*self.msh.J
        self.fld.re = self.fld.re*self.msh.J

    def log(self):
        """ Display timings. """

        if self.cfg.it % self.cfg.ns == 0:
            self.save()
            if self.cfg.timings and not self.cfg.quiet:
                self.bench['total'].append(time.perf_counter() - self.tt)
                self.bench = timing.disp(self.bench, self.cfg.it, self.res)
        elif self.cfg.timings and not self.cfg.quiet:
            self.bench['total'].append(time.perf_counter() - self.tt)

    def __str__(self):

        rep = '-'*int(self.columns)
        rep += '\n'
        rep += 'Mesh shape    : {}\n'.format(self.msh.shape)
        rep += 'Spacial steps : dx={}, dz={}\n'.format(self.msh.dx, self.msh.dz)
        rep += 'CFL           : {}\n'.format(self.cfg.CFL)
        rep += 'Time step     : {}\n'.format(self.cfg.dt)
        rep += '-'*int(self.columns)
        return rep

    def __repr__(self):
        return self.__str__()

    @property
    def columns(self):
        """ Return max terminal width. """
        try:
            _, col = os.popen('stty size', 'r').read().split()
            return int(col) if int(col) < 81 else 80
        except ValueError:
            return 80



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : 2016-11-29 - 23:18:27
#
# pylint: disable=too-many-statements
# pylint: disable=attribute-defined-outside-init
# pylint: disable=too-many-instance-attributes
"""
-----------

This module contains the :py:class:`CfgSetup` that read the configuration file
and set all simulation parameters.

Example
-------

::

    from nsfds2.init import CfgSetup

    cfg = CfgSetup()

-----------
"""

import json
import time
import sys
import os
import shutil
import datetime
import pathlib
import configparser
from pkg_resources import parse_version
import nsfds2
from nsfds2.utils import files


class CfgSetup:
    """ Handle configuration file. """

    def __init__(self, args=None):

        # Minimal version of the config file
        self.base_version = '0.9.15'

        # Command line arguments + home
        self.args = args
        self.home = pathlib.Path.home()
        self.path_default = self.home / '.nsfds2'
        self.cfgfile_default = self.path_default / 'nsfds2.conf'

        # Create config parser
        self.cfg = configparser.ConfigParser(allow_no_value=True)

        # Load config file
        if isinstance(self.args, str):
            self.cfgfile = pathlib.Path(self.args)
        else:
            self.cfgfile = getattr(self.args, 'cfgfile', None)

        # Check cfg file
        if not self.cfgfile:
            self.path = self.path_default
            self.cfgfile = self.cfgfile_default
            self.check_dir(self.path) # Check if cfg dir exists. If not create it.
            self.init_cfg()           # Check if cfg file exist. If not create it
        else:
            self.cfgfile = pathlib.Path(self.cfgfile)
            self.path = self.cfgfile.absolute().parent

        # Check if config file version is ok
        self.check_config_file()

        # Read config file (can be overridden by command line)
        self.cfg.read(self.cfgfile)

        # Parse arguments
        self.run()

    @staticmethod
    def check_dir(directory):
        """ Check if dir exists. If not, create it."""

        if not directory.is_dir():
            directory.mkdir()
            print("Create directory :", directory)
            time.sleep(0.5)

    def check_config_file(self):
        """ Check version of the config file. Overwrite it if too old. """

        cfg = configparser.ConfigParser(allow_no_value=True)
        cfg.read(self.cfgfile)

        try:
            CFG = cfg['configuration']
            version = CFG.get('version')
            is_default_cfg = self.cfgfile == self.cfgfile_default
            is_version_ok = parse_version(version) >= parse_version(self.base_version)

            if not is_version_ok and is_default_cfg:
                self._overwrite_config_file()
            elif not is_version_ok:
                print(f'Config file version must be >= {self.base_version}')
                sys.exit(1)

        except (KeyError, TypeError):
            print(f'Config file version must be >= {self.base_version}')
            if is_default_cfg:
                self._overwrite_config_file()
            else:
                sys.exit(1)

    def _overwrite_config_file(self):

        # Backup old config file
        now = datetime.datetime.now()
        name = f'{now.year}{now.month}{now.day}{now.hour}{now.minute}{now.second}'
        shutil.move(self.path / 'nsfds2.conf', self.path / f'nsfds2_{name}.conf')

        print(f'Current configfile backup : nsfds2_{name}.conf')
        time.sleep(1)

        # Create new config file
        self.init_cfg()

    def init_cfg(self):
        """ Check if config file exists. If not create it. """

        if not (self.path / 'nsfds2.conf').is_file():
            open(self.path / 'nsfds2.conf', 'a').close()
            print("Create configuration file : {}/nsfds2.conf".format(self.path))
            time.sleep(0.5)
            self.write_default()

    def write_default(self):
        """ Write default configuration file. """

        self.cfg.add_section('configuration')
        self.cfg.set('configuration', 'version', str(nsfds2.__version__))
        self.cfg.set('configuration', 'timings', 'False')
        self.cfg.set('configuration', 'quiet', 'False')
        self.cfg.set('configuration', 'cpu', '1')

        self.cfg.add_section('simulation')
        self.cfg.set('simulation', 'nt', '500')
        self.cfg.set('simulation', 'ns', '10')
        self.cfg.set('simulation', 'CFL', '0.5')

        self.cfg.add_section('thermophysic')
        self.cfg.set('thermophysic', 'P0', '101325.0')
        self.cfg.set('thermophysic', 'T0', '20.0')
        self.cfg.set('thermophysic', 'gamma', '1.4')
        self.cfg.set('thermophysic', 'prandtl', '0.7')

        self.cfg.add_section('geometry')
        self.cfg.set('geometry', 'mesh', 'regular')
        self.cfg.set('geometry', 'file', 'None')
        self.cfg.set('geometry', 'geoname', 'helmholtz_double')
        self.cfg.set('geometry', 'curvname', 'None')
        self.cfg.set('geometry', 'bc', 'PPPP')
        self.cfg.set('geometry', 'nx', '256')
        self.cfg.set('geometry', 'nz', '256')
        self.cfg.set('geometry', 'ix0', '0')
        self.cfg.set('geometry', 'iz0', '0')
        self.cfg.set('geometry', 'dx', '1')
        self.cfg.set('geometry', 'dz', '1')

        self.cfg.add_section('PML')
        self.cfg.set('PML', 'beta', '0.')
        self.cfg.set('PML', 'alpha', '4.')
        self.cfg.set('PML', 'sigmax', 'auto')
        self.cfg.set('PML', 'sigmaz', 'auto')
        self.cfg.set('PML', 'Npml', '15')

        self.cfg.add_section('source')
        self.cfg.set('source', 'type', 'pulse')
        self.cfg.set('source', 'ixS', '32')
        self.cfg.set('source', 'izS', '128')
        self.cfg.set('source', 'S0', '1e6')
        self.cfg.set('source', 'B0', '5')
        self.cfg.set('source', 'f0', '20000')
        self.cfg.set('source', 'wavfile', 'None')

        self.cfg.add_section('flow')
        self.cfg.set('flow', 'type', 'None')
        self.cfg.set('flow', 'U0', '5')
        self.cfg.set('flow', 'V0', '5')

        self.cfg.add_section('eulerian fluxes')
        self.cfg.set('eulerian fluxes', 'stencil', '11')

        self.cfg.add_section('filtering')
        self.cfg.set('filtering', 'filter', 'True')
        self.cfg.set('filtering', 'stencil', '11')
        self.cfg.set('filtering', 'stength', '0.75')

        self.cfg.add_section('viscous fluxes')
        self.cfg.set('viscous fluxes', 'viscosity', 'True')
        self.cfg.set('viscous fluxes', 'stencil', '7')

        self.cfg.add_section('shock capture')
        self.cfg.set('shock capture', 'shock capture', 'True')
        self.cfg.set('shock capture', 'stencil', '7')
        self.cfg.set('shock capture', 'method', 'pressure')

        self.cfg.add_section('figures')
        self.cfg.set('figures', 'figures', 'True')
        self.cfg.set('figures', 'probes', 'True')
        self.cfg.set('figures', 'pml', 'True')
        self.cfg.set('figures', 'bc_profiles', 'True')
        self.cfg.set('figures', 'fps', '24')

        self.cfg.add_section('save')
        self.cfg.set('save', 'path', 'results/')
        self.cfg.set('save', 'filename', 'tmp')
        self.cfg.set('save', 'compression', 'lzf')
        self.cfg.set('save', 'fields', 'True')
        self.cfg.set('save', 'probes', '[]')

        with open(self.path / 'nsfds2.conf', 'w') as cf:
            self.cfg.write(cf)

    def run(self):
        """ Run configuration. """

        self.none = ['none', 'false', '']

        try:
            self._cfg()
            self._sim()
            self._thp()
            self._geo()
            self._pml()
            self._src()
            self._flw()
            self._eul()
            self._flt()
            self._vsc()
            self._cpt()
            self._save()
            self._figs()

        except configparser.Error as err:
            print('Bad cfg file : ', err)
            sys.exit(1)

        self.dt = min(self.dx, self.dz)*self.CFL/(self.c0 +
                                                  max(abs(self.U0), abs(self.V0)))

    def _cfg(self):

        CFG = self.cfg['configuration']
        self.timings = getattr(self.args, 'timings', None)
        self.quiet = getattr(self.args, 'quiet', None)
        self.cpu = CFG.getint('cpu', 1)

        if not isinstance(self.timings, bool):
            self.timings = CFG.getboolean('timings', False)

        if not isinstance(self.quiet, bool):
            self.quiet = CFG.getboolean('quiet', False)

    def _sim(self):

        SIM = self.cfg['simulation']
        self.nt = getattr(self.args, 'nt', None)
        self.ns = SIM.getint('ns', 10)
        self.CFL = SIM.getfloat('CFL', 0.5)

        if self.nt is None:
            self.nt = SIM.getint('nt', 500)

        if self.nt % self.ns:
            self.nt -= self.nt % self.ns

        self.it = 0

    def _thp(self):

        THP = self.cfg['thermophysic']

        self.Ssu = 111.0  # Sutherland constant
        self.T0 = 273.0
        self.T = self.T0 + THP.getfloat('T0', 20.0)

        self.gamma = THP.getfloat('gamma', 1.4)
        self.cv = 717.5
        self.cp = self.cv*self.gamma

        self.p0 = THP.getfloat('P0', 101325.0)
        self.rho0 = self.p0/(self.T*(self.cp - self.cv))
        self.c0 = (self.gamma*self.p0/self.rho0)**0.5
        self.mu0 = 0.00001716
        self.mu = (self.mu0*(self.T/self.T0)**(3./2.) *
                   (self.T0 + self.Ssu)/(self.T + self.Ssu))
        self.nu = self.mu/self.rho0
        self.prandtl = THP.getfloat('prandtl', 0.7)

#        self.nu = THP.getfloat('nu', 1.5e-5)
#        self.p0 = self.rho0*self.c0**2/self.gamma

        if self.c0 < 1:
            raise ValueError('c0 must be >= 1')

    def _geo(self):

        GEO = self.cfg['geometry']
        self.mesh = GEO.get('mesh', 'regular').lower()
        self.curvflag = True if self.mesh == 'curvilinear' else False
        self.curvname = GEO.get('curvname', 'None')
        self.geofile = GEO.get('file', 'None')
        self.geoname = GEO.get('geoname', 'square')
        self.geoflag = True
        self.bc = GEO.get('bc', 'WWWW').upper()
        self.nx = GEO.getint('nx', 256)
        self.nz = GEO.getint('nz', 256)
        self.ix0 = GEO.getint('ix0', 0)
        self.iz0 = GEO.getint('iz0', 0)
        self.dx = GEO.getfloat('dx', 1)
        self.dz = GEO.getfloat('dz', 1)

        if self.geofile != "None":
            self.geofile = self.path / self.geofile

        if self.mesh not in ['regular', 'adaptative', 'curvilinear']:
            raise ValueError('mesh must be regular, adaptative, or curvilinear')

        self.obstacles = files.get_obstacle(self)

    def _pml(self):

        PML = self.cfg['PML']
        self.beta = PML.getfloat('beta', 0.)
        self.alpha = PML.getfloat('alpha', 4.)
        self.sigmax = PML.get('sigmax', 'auto')
        self.sigmaz = PML.get('sigmaz', 'auto')
        self.Npml = PML.getint('Npml', 15)

    def _src(self):

        SRC = self.cfg['source']
        self.stype = SRC.get('type', 'pulse').lower()
        self.ixS = SRC.getint('ixS', 32)
        self.izS = SRC.getint('izS', 32)
        self.S0 = SRC.getfloat('S0', 1e3)
        self.B0 = SRC.getfloat('B0', 5)
        self.f0 = SRC.getfloat('f0', 20000)
        self.wavfile = SRC.get('wavfile', None)
        self.seed = SRC.get('seed', None)
        self.off = SRC.getint('off', self.nt)

        if self.wavfile:
            self.wavfile = pathlib.Path(self.wavfile).expanduser()

        if self.seed:
            try:
                self.seed = int(self.seed)
            except ValueError:
                raise ValueError('Seed must be int or None')

        if self.stype in self.none:
            self.S0 = 0

    def _flw(self):

        FLW = self.cfg['flow']
        self.ftype = FLW.get('type', 'None').lower()
        self.U0 = FLW.getfloat('U0', 5)
        self.V0 = FLW.getfloat('V0', 5)

        if self.ftype in self.none:
            self.U0, self.V0 = 0., 0.

    def _eul(self):

        EUL = self.cfg['eulerian fluxes']
        self.stencil = EUL.getint('stencil', 11)

        if self.stencil not in [3, 7, 11]:
            raise ValueError('stencil must be 3, 7 or 11')

    def _flt(self):

        FLT = self.cfg['filtering']
        self.flt = FLT.getboolean('filter', True)
        self.flt_stencil = FLT.getint('stencil', 11)
        self.xnu = FLT.getfloat('strength', 0.75)

        if self.flt_stencil not in [7, 11]:
            raise ValueError('only 7 and 11 pts filters implemented for now')

    def _vsc(self):

        VSC = self.cfg['viscous fluxes']
        self.vsc = VSC.getboolean('viscosity', True)
        self.vsc_stencil = VSC.getint('stencil', 3)

        if self.vsc_stencil not in [3, 7, 11]:
            raise ValueError('viscous fluxes only available with 3, 7 or 11 pts')

    def _cpt(self):

        CPT = self.cfg['shock capture']
        self.cpt = CPT.getboolean('shock capture', True)
        self.cpt_stencil = CPT.getint('stencil', 7)
        self.cpt_meth = CPT.get('method', 'pressure').lower()
        self.rth = 1e-6

        if self.cpt_stencil not in [3, 7, 11]:
            raise ValueError('capture only available with 3, 7 or 11 pts')

        if self.cpt_meth not in ['pressure', 'dilatation']:
            raise ValueError('capture method must be pressure or dilatation')

    def _save(self):

        SAVE = self.cfg['save']
        self.save = SAVE.getboolean('fields', True)
        if self.path == self.path_default:
            self.savepath = pathlib.Path(SAVE.get('path', 'results'))
        else:
            self.savepath = self.path / SAVE.get('path', 'results')
        self.savefile = SAVE.get('filename', 'tmp') + '.hdf5'
        self.comp = SAVE.get('compression', 'lzf')
        try:
            self.probes = json.loads(SAVE.get('probes', '[]'))
        except:
            raise ValueError('probe and probe_location have been merged. Update config file.')

        # Check probes
        if self.probes:
            for c in self.probes:
                if not 0 <= c[0] < self.nx or not 0 <= c[1] < self.nz:
                    raise ValueError('probes must be in the domain')

        # datapath and datafile
        if self.comp == 'None':
            self.comp = None

        # if self.savepath does not exist, create it
        self.check_dir(self.savepath)

        self.datafile = getattr(self.args, 'datafile', None)
        if not self.datafile:
            self.datafile = self.savepath / self.savefile
        else:
            self.datafile = pathlib.Path(self.datafile).expanduser()

    def _figs(self):

        FIGS = self.cfg['figures']
        self.figures = FIGS.getboolean('figures', True)
        self.show_probes = FIGS.getboolean('probes', True)
        self.show_pml = FIGS.getboolean('pml', True)
        self.bc_profiles = FIGS.getboolean('bc_profiles', True)
        self.fps = FIGS.getint('fps', 24)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
# Creation Date : 2019-04-19 - 15:14:22
"""
-----------

Utils

-----------
"""

import re
import warnings


class Check:
    """ Check FDTD parameters. """

    def __init__(self, cfg, msh):

        self.cfg = cfg
        self.msh = msh

        self.source()
        self.obstacles()
        self.domain()
        self.not_implemented()

    def source(self):
        """ Check if source is not in an obstacle. """
        for o in self.msh.obstacles:
            if o.ix[0] < self.cfg.ixS < o.ix[1] and o.iz[0] < self.cfg.izS < o.iz[1]:
                raise ValueError('source cannot be in an obstacle')

    def obstacles(self):
        """ Check validity of obstacles boundary conditions. """

        for obs in self.msh.obstacles:
            if not re.match(r'^[WV]+$', obs.bc):
                s = "Obstacles can only be combination of 'WV' for now. "
                raise ValueError(s)

    def domain(self):
        """ Check validity of the bcs. """

        if not re.match(r'[PRAW][PRAW][PRAW][PRAW]', self.msh.bc):
            s = "Only 'W', 'R', 'P', and 'A' bc are implemented for now. "
            s += "Fix bcs to 'WWWW'."
            warnings.warn(s, stacklevel=8)
            self.msh.bc = 'WWWW'

    def not_implemented(self):
        """ Check not implemented features. """

        if self.cfg.mesh == "curvilinear" and self.cfg.vsc:
            s = 'Cannot solve viscous fluxes with curvilinear mesh for now'
            raise ValueError(s)




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
# Creation Date : 2019-03-08 - 10:47:50
#
# pylint: disable=redefined-builtin
"""
-----------

Headers for nsfds2

-----------
"""

import os
import sys
import platform
import textwrap
import numpy
import ofdlib2
import fdgrid
import nsfds2
import datetime


def _columns():
    try:
        _, col = os.popen('stty size', 'r').read().split()
        return int(col) if int(col) < 81 else 80
    except ValueError:
        return 80


def copyright():
    """ Show copyright. """
    col = 76 if _columns() > 76 else _columns
    cp = [col*'#',
          f"nsfds2 v{nsfds2.__version__} -- Copyright (C) 2016-{datetime.datetime.now().year} -- Cyril Desjouy",
          " "]
    lc = "This program comes with ABSOLUTELY NO WARRANTY. " + \
         "This is free software, and you are welcome to redistribute it " + \
         "under certain conditions; See GNU GPL v3 for more details."
    cp += textwrap.wrap(lc, int(col))
    cp += [col*"#"]
    tmp = ''
    for txt in cp:
        tmp += "# {:^{col}} #\n".format(txt, col=col)
    print(tmp)


def version():
    """ Display versions of modules in use. """

    print('System : {}\n'.format(platform.platform()))
    print('\tpython : {:<10}\tnumpy  : {:<10}'.format(sys.version.split(' ')[0], numpy.__version__))
    print('\tfdgrid : {:<10}\tofdlib : {:<10}'.format(fdgrid.__version__, ofdlib2.__version__))
    print('\n' + _columns()*"#" + "\n")


def start(cfg):
    """ Prompt start. """

    check_geo(cfg)

    log = "Starting computation for geometry : '{}' ({}x{})."
    log += "\nSolving {} equations {} shock capture procedure."
    eqs = "Navier-Stokes" if cfg.vsc else "Nonlinear Euler"
    cpt = "with" if cfg.cpt else "without"
    print(log.format(cfg.geoname, cfg.nx, cfg.nz, eqs, cpt))
    key = input("Hit enter to continue (prefix 'q' to abort) ! ")
    if key == 'q':
        sys.exit(0)


def check_geo(cfg):
    """ Check geofile. """

    if not cfg.geoflag:
        print('Warning: Unable to load {} from {}.py'.format(cfg.geoname, cfg.geofile))
        cfg.geoname = 'empty'


def parameters(cfg, msh):
    """ Show simulation parameters. """

    # Geometry
    s = msh.__str__()
    s += "\t* geometry     : '{}'\n".format(cfg.geoname)

    # Thermophysics
    s += "\t* Thermophysic : "
    s += "c0={:.2f} m/s, rho0={:.2f} kg/m3, nu={:.3e} m2/s\n".format(cfg.c0,
                                                                     cfg.rho0,
                                                                     cfg.nu)
    # PML
    if 'A' in cfg.bc:
        nsfds2.init.fields.set_sigma(cfg, msh)
        s += "\t* sigma=({:.3e}, {:.3e}) and beta={}\n".format(cfg.sigmax,
                                                               cfg.sigmaz,
                                                               cfg.beta)

    # Time
    s += "\t* Physical time: {:.5e} s.\n".format(cfg.dt*cfg.nt)
    s += "\t* Time step    : dt={:.5e} s and nt={}.\n".format(cfg.dt,
                                                              cfg.nt)

    # Source
    if cfg.stype not in cfg.none:
        s += "\t* source   : {} at ({}, {})".format(cfg.stype,
                                                    cfg.ixS,
                                                    cfg.izS)
        if cfg.ftype == 'harmonic':
            s += 'with f={}.\n'.format(cfg.f0)
        else:
            s += '.\n'

    if cfg.ftype not in cfg.none:
        s += "\t* flow     : {} (U0={}, V0={}).".format(cfg.ftype,
                                                        cfg.U0,
                                                        cfg.V0)

    print(s)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : 2019-03-01 - 14:09:42
#
# pylint: disable=too-many-instance-attributes
"""
-----------

To initialize all the fields, :py:class:`Fields` needs :

    * the mesh (from :py:class:fdgrid.mesh.Mesh)
    * the configuration of the simulation as an instance of :py:class:`nsfds2.init.config.CfgSetup`


Example
-------

::

    fld = Fields(msh, cfg)

-----------
"""

import sys as _sys
import pickle as _pickle
import h5py as _h5py
import numpy as _np
from ofdlib2.fdtd import fdtools
from nsfds2.utils import headers, misc
from nsfds2.utils.sounds import resample
from nsfds2.utils.files import get_wall_function


def set_sigma(cfg, msh):
    """ Set sigmax and sigmaz for PMLs. """

    try:
        cfg.sigmax = float(cfg.sigmax)
    except ValueError:
        D = msh.x[msh.Npml] - msh.x[0]
        cfg.sigmax = 8*cfg.c0*(cfg.alpha+1)/D

    try:
        cfg.sigmaz = float(cfg.sigmaz)
    except ValueError:
        D = msh.z[msh.Npml] - msh.z[0]
        cfg.sigmaz = 8*cfg.c0*(cfg.alpha+1)/D


class Fields:
    """ Initialize fields.

    Parameters
    ----------
    msh : :py:class:fdgrid.mesh.Mesh
        The mesh
    cfg : :py:class:`nsfds2.init.config.CfgSetup`
        The simulation configuration

    """

    def __init__(self, msh, cfg):

        self._msh = msh
        self._cfg = cfg

        self._nx = self._cfg.nx
        self._nz = self._cfg.nz
        self._x = self._msh.x
        self._z = self._msh.z

        # Headers
        if not self._cfg.quiet:
            headers.copyright()
            headers.version()

        # Fields
        self._init_fields()
        self._init_derivatives()
        self._init_filters()
        self._init_wall_sources()

        self.fdtools = fdtools(self._nx, self._nz,
                               self._cfg.p0, self._cfg.gamma, self._cfg.dt)

        # PMLS
        if 'A' in self._msh.bc:
            self._init_pml()

        if 'R' in self._msh.bc:
            self._init_radiation()

        # Obstacles to 0
        self.zero_obstacles()

    def _init_fields(self):
        """ Setup initial fields. """

        if self._cfg.mesh == 'curvilinear':
            self.p = _np.zeros(self._msh.shape) + self._cfg.p0/self._msh.J
            self.r = _np.zeros_like(self.p) + self._cfg.rho0/self._msh.J
            self.ru = _np.zeros_like(self.p) + self._cfg.U0/self._msh.J
            self.rv = _np.zeros_like(self.p) + self._cfg.V0/self._msh.J
        else:
            self.p = _np.zeros(self._msh.shape) + self._cfg.p0
            self.r = _np.zeros_like(self.p) + self._cfg.rho0
            self.ru = _np.zeros_like(self.p) + self._cfg.U0
            self.rv = _np.zeros_like(self.p) + self._cfg.V0

        if self._cfg.cpt_meth == 'dilatation':
            self.dltn = _np.zeros_like(self.p)

        # Source
        self.source_select()

        # Flow
        self.flow_select()

        # Update rho.e
        self.re = self.p/(self._cfg.gamma-1.)

    def _init_derivatives(self):
        """ Init derivatives. """

        self.K = _np.zeros_like(self.p)
        self.Ku = _np.zeros_like(self.p)
        self.Kv = _np.zeros_like(self.p)
        self.Ke = _np.zeros_like(self.p)

        self.E = _np.empty_like(self.p)
        self.Eu = _np.empty_like(self.p)
        self.Ev = _np.empty_like(self.p)
        self.Ee = _np.empty_like(self.p)

        self.F = _np.empty_like(self.p)
        self.Fu = _np.empty_like(self.p)
        self.Fv = _np.empty_like(self.p)
        self.Fe = _np.empty_like(self.p)

    def _init_filters(self):
        """ Init variables used for filtering. """

        self.dp = _np.zeros_like(self.p)
        self.sg = _np.zeros_like(self.p)
        self.tau11 = _np.zeros_like(self.p)
        self.tau22 = _np.zeros_like(self.p)
        self.tau12 = _np.zeros_like(self.p)

    def _init_pml(self):
        """ Init PMLs. """

        set_sigma(self._cfg, self._msh)

        # sigmax
        Dx = self._x[self._msh.Npml] - self._x[0]             # Width of the PML
        self.sx = _np.zeros(self._nx)
        self.sx[:self._msh.Npml] = self._cfg.sigmax*abs((self._x[:self._msh.Npml] \
                                 - self._x[self._msh.Npml])/Dx)**self._cfg.alpha
        self.sx[self._nx - self._msh.Npml:] = self.sx[self._msh.Npml-1::-1]

        # sigmaz
        Dz = self._z[self._msh.Npml] - self._z[0]             # Width of the PML
        self.sz = _np.zeros(self._nz)
        self.sz[:self._msh.Npml] = self._cfg.sigmaz*abs((self._z[:self._msh.Npml] \
                                 - self._z[self._msh.Npml])/Dz)**self._cfg.alpha
        self.sz[self._nz - self._msh.Npml:] = self.sz[self._msh.Npml-1::-1]

        # Init Q
        self.qx = _np.zeros_like(self.p)
        self.qux = _np.zeros_like(self.p)
        self.qvx = _np.zeros_like(self.p)
        self.qex = _np.zeros_like(self.p)
        self.qz = _np.zeros_like(self.p)
        self.quz = _np.zeros_like(self.p)
        self.qvz = _np.zeros_like(self.p)
        self.qez = _np.zeros_like(self.p)

        # Init K
        self.Kx = _np.zeros_like(self.p)
        self.Kux = _np.zeros_like(self.p)
        self.Kvx = _np.zeros_like(self.p)
        self.Kex = _np.zeros_like(self.p)
        self.Kz = _np.zeros_like(self.p)
        self.Kuz = _np.zeros_like(self.p)
        self.Kvz = _np.zeros_like(self.p)
        self.Kez = _np.zeros_like(self.p)

        # Initial E & F
        self.Ei = _np.zeros_like(self.p)
        self.Eui = _np.zeros_like(self.p)
        self.Evi = _np.zeros_like(self.p)
        self.Eei = _np.zeros_like(self.p)
        self.Fi = _np.zeros_like(self.p)
        self.Fui = _np.zeros_like(self.p)
        self.Fvi = _np.zeros_like(self.p)
        self.Fei = _np.zeros_like(self.p)

        if self._cfg.mesh in ['regular', 'adaptative']:
            self.fdtools.Eu(self.Ei, self.Eui, self.Evi, self.Eei,
                            self.r, self.ru, self.rv,
                            self._cfg.p0*_np.ones_like(self.p)/(self._cfg.gamma-1.),
                            self._cfg.p0*_np.ones_like(self.p))

        elif self._cfg.mesh == 'curvilinear':
            self.fdtools.EuJ(self.Ei, self.Eui, self.Evi, self.Eei,
                             self.r, self.ru, self.rv, self.re, self.p,
                             self._msh.dxn_dxp, self._msh.dxn_dzp)

        if self._cfg.mesh in ['regular', 'adaptative']:
            self.fdtools.Fu(self.Fi, self.Fui, self.Fvi, self.Fei,
                            self.r, self.ru, self.rv,
                            self._cfg.p0*_np.ones_like(self.p)/(self._cfg.gamma-1.),
                            self._cfg.p0*_np.ones_like(self.p))

        elif self._cfg.mesh == 'curvilinear':
            self.fdtools.FuJ(self.Fi, self.Fui, self.Fvi, self.Fei,
                             self.r, self.ru, self.rv, self.re, self.p,
                             self._msh.dzn_dxp, self._msh.dzn_dzp)

    def _init_radiation(self):
        """ Init radiation conditions. """

        # Init K
        self.Kx = _np.zeros_like(self.p)
        self.Kux = _np.zeros_like(self.p)
        self.Kvx = _np.zeros_like(self.p)
        self.Kex = _np.zeros_like(self.p)
        self.Kz = _np.zeros_like(self.p)
        self.Kuz = _np.zeros_like(self.p)
        self.Kvz = _np.zeros_like(self.p)
        self.Kez = _np.zeros_like(self.p)

        xCL = self._x[self._cfg.ixS]
        zCL = self._z[self._cfg.izS]

        self.ixCL = _np.zeros(11)
        self.izCL = _np.zeros(11)

        self.ixCL[-5] = self._nx - 5
        self.ixCL[-4] = self._nx - 4
        self.ixCL[-3] = self._nx - 3
        self.ixCL[-2] = self._nx - 2
        self.ixCL[-1] = self._nx - 1
        self.ixCL[0] = 1
        self.ixCL[1] = 2
        self.ixCL[2] = 3
        self.ixCL[3] = 4
        self.ixCL[4] = 5

        self.izCL[-5] = self._nz - 5
        self.izCL[-4] = self._nz - 4
        self.izCL[-3] = self._nz - 3
        self.izCL[-2] = self._nz - 2
        self.izCL[-1] = self._nz - 1
        self.izCL[0] = 0
        self.izCL[1] = 1
        self.izCL[2] = 2
        self.izCL[3] = 3
        self.izCL[4] = 4

        self.r_v = _np.zeros((10, self._nz))
        self.cosv = _np.zeros((10, self._nz))
        self.sinv = _np.zeros((10, self._nz))
        self.r_h = _np.zeros((self._nx, 10))
        self.cosh = _np.zeros((self._nx, 10))
        self.sinh = _np.zeros((self._nx, 10))


        # prendre (xCL,yCL) pour origine pour le calcul de one_r, cosphi et sinphi
        # calcul pour les bandes verticales
        for ix in range(-5, 5):
            self.r_v[ix, :] = 0.5/_np.sqrt((self._x[ix] - xCL)**2 + (self._z - zCL)**2)
            self.cosv[ix, :] = (self._x[ix] - xCL)*2*self.r_v[ix, :]
            self.sinv[ix, :] = (self._z - zCL)*2*self.r_v[ix, :]

        # calcul pour les bandes horizontales
        for iz in range(-5, 5):
            self.r_h[:, iz] = 0.5/_np.sqrt((self._x - xCL)**2 + (self._z[iz] - zCL)**2)
            self.cosh[:, iz] = (self._x - xCL)*2.*self.r_h[:, iz]
            self.sinh[:, iz] = (self._z[iz] - zCL)*2.*self.r_h[:, iz]

    def _init_wall_sources(self):
        """ Init all sources on walls. """

        time = _np.arange(0, (self._cfg.nt+1)*self._cfg.dt, self._cfg.dt)

        for obs in self._msh.obstacles:
            for edge in obs.edges:
                if edge.f0 > 0:
                    edge.vn = _np.sin(2*_np.pi*edge.f0*time + edge.phi)
                elif edge.func:
                    func = get_wall_function(self._cfg, edge.func)
                    edge.vn = func(time)
                elif edge.wav:
                    filename = edge.wav.replace('~', str(self._cfg.home))
                    if filename in self.wav_lst:
                        edge.vn = self.wav_lst[filename]
                    else:
                        edge.vn = resample(filename, 1/self._cfg.dt,
                                           pad=self._cfg.nt+1)
                        self.wav_lst[filename] = edge.vn
                else:
                    edge.vn = 0


    def init_save(self):
        """ Init save. """

        if self._cfg.datafile.is_file():
            msg = f'{self._cfg.datafile} already exists. \nOverwrite ? [yes]/no '
            overwrite = input(misc.colors.RED + msg + misc.colors.END)
            if overwrite.lower() in ['n', 'no']:
                _sys.exit(1)

        with open(self._cfg.datafile.with_suffix('.cfg'), 'wb') as pkl:
            _pickle.dump(self._cfg, pkl)

        self.sfile = _h5py.File(self._cfg.datafile, 'w')

        self.sfile.create_dataset('x', data=self._x, compression=self._cfg.comp)
        self.sfile.create_dataset('z', data=self._z, compression=self._cfg.comp)
        self.sfile.create_dataset('rho_init', data=self.r, compression=self._cfg.comp)
        self.sfile.create_dataset('rhou_init', data=self.ru, compression=self._cfg.comp)
        self.sfile.create_dataset('rhov_init', data=self.rv, compression=self._cfg.comp)
        self.sfile.create_dataset('rhoe_init', data=self.re, compression=self._cfg.comp)

        self.sfile.attrs['obstacles'] = self._msh.get_obstacles()
#        self.sfile.attrs['domains'] = self._msh.get_domains(only_xz=True)
        self.sfile.attrs['dx'] = self._msh.dx
        self.sfile.attrs['dz'] = self._msh.dz
        self.sfile.attrs['dt'] = self._cfg.dt
        self.sfile.attrs['nx'] = self._nx
        self.sfile.attrs['nz'] = self._nz
        self.sfile.attrs['nt'] = self._cfg.nt
        self.sfile.attrs['ns'] = self._cfg.ns
        self.sfile.attrs['p0'] = self._cfg.p0
        self.sfile.attrs['rho0'] = self._cfg.rho0
        self.sfile.attrs['gamma'] = self._cfg.gamma
        self.sfile.attrs['Npml'] = self._cfg.Npml
        self.sfile.attrs['mesh'] = self._cfg.mesh
        self.sfile.attrs['bc'] = self._cfg.bc

        probes = _np.zeros((len(self._cfg.probes), self._cfg.nt))
        self.sfile.create_dataset('probes_location', data=self._cfg.probes)
        self.sfile.create_dataset('probes_value', data=probes,
                                  compression=self._cfg.comp)

        if self._cfg.mesh == 'curvilinear':
            self.sfile.create_dataset('J', data=self._msh.J, compression=self._cfg.comp)
            self.sfile.create_dataset('xn', data=self._msh.xn, compression=self._cfg.comp)
            self.sfile.create_dataset('zn', data=self._msh.zn, compression=self._cfg.comp)
            self.sfile.create_dataset('xp', data=self._msh.xp, compression=self._cfg.comp)
            self.sfile.create_dataset('zp', data=self._msh.zp, compression=self._cfg.comp)

    def source_select(self):
        """ Source selection. """

        self.Bx = self._cfg.B0*self._cfg.dx
        self.src = _np.empty_like(self.p)
        self.update_source = None
        self.wav_lst = dict()

        if self._cfg.stype in ["", "None", "none"]:
            pass

        elif self._cfg.stype == "pulse":
            self.pulse()

        elif self._cfg.stype == "harmonic":
            self.update_source = self.update_harmonic
            self.harmonic()

        elif self._cfg.stype == "white":
            self.update_source = self.update_white
            self.white()

        elif self._cfg.stype == "wav" and self._cfg.wavfile:
            self.wav = resample(self._cfg.wavfile, 1/self._cfg.dt, pad=self._cfg.nt+1)
            self.wav_lst[self._cfg.wavfile] = self.wav
            self.update_source = self.update_wav
            self.harmonic()

        else:
            raise ValueError('Only pulse, harmonic, white, wave supported for now')

    def flow_select(self):
        """ Flow selection. """

        if self._cfg.ftype in ["", "None", "none"]:
            pass
        elif self._cfg.ftype == "custom":
            pass
        elif self._cfg.ftype == "vortex":
            self.isentropic_vortex()
        elif self._cfg.ftype == "poiseuille":
            self.poiseuille()
        elif self._cfg.ftype == "kh":
            self.kelvin_helmholtz()
        else:
            raise ValueError('Only custom and vortex supported for now')

    def pulse(self):
        """ Pulse source. """

        # Location
        ixS = self._cfg.ixS
        izS = self._cfg.izS

        if self._cfg.mesh == 'curvilinear':
            for iz in range(self._nz):
                for ix in range(self._nx):
                    self.p[ix, iz] = self._cfg.p0/self._msh.J[ix, iz] \
                             + self._cfg.S0*_np.exp(-_np.log(2) \
                             * ((self._msh.xp[ix, iz] - self._msh.xp[ixS, izS])**2 +
                                (self._msh.zp[ix, iz] - self._msh.zp[ixS, izS])**2)/self.Bx**2)
        else:
            for iz, z in enumerate(self._z):
                self.p[:, iz] += self._cfg.S0*_np.exp(-_np.log(2) \
                            * ((self._x-self._x[ixS])**2 +
                               (z - self._z[izS])**2)/self.Bx**2)

        # update rho
        self.r = self.p/(self._cfg.c0*self._cfg.c0)

    def harmonic(self):
        """ Harmonic ponctual source. """

        for iz, z in enumerate(self._z):
            self.src[:, iz] = _np.exp(-_np.log(2)*((self._x - self._x[self._cfg.ixS])**2 +
                                                   (z - self._z[self._cfg.izS])**2)/self.Bx**2)
        self.src = self._cfg.S0*self.src

    def white(self):
        """ white noise. """

        for iz, z in enumerate(self._z):
            tmp = (1 - _np.log(2)*((self._x - self._x[self._cfg.ixS])**2 +
                                   (z - self._z[self._cfg.izS])**2)/self.Bx**2)
            self.src[:, iz] = tmp*_np.exp(-_np.log(2)*((self._x - self._x[self._cfg.ixS])**2 +
                                                       (z - self._z[self._cfg.izS])**2)/self.Bx**2)
        self.src = self._cfg.S0*self.src
        _np.random.seed(self._cfg.seed)
        self.noise = _np.random.normal(size=self._cfg.nt)

    def update_harmonic(self, it):
        """ Harmonic source time evolution. """

        if it <= self._cfg.off:
            return self.src*_np.sin(2*_np.pi*self._cfg.f0*it*self._cfg.dt)

        return 0

    def update_white(self, it):
        """ White noise time evolution. """

        if it <= self._cfg.off:
            return self.src*self.noise[it-1]

        return 0

    def isentropic_vortex(self):
        """ Initialize fields with an isentropic vortex [Hu, JCP, 2008] """

        Umax = 0.5*self._cfg.U0
        b = 0.2

        theta = _np.zeros(self._nx)

        # initialisation des fluctuations
        for iz in range(self._nz):

            r = _np.sqrt(self._x**2 + self._z[iz]**2)

            for ix in range(self._nx):
                if self._x[ix] == 0 and self._z[iz] == 0:
                    theta[ix] = 0
                elif self._x[ix] >= 0:
                    theta[ix] = _np.arcsin(self._z[iz]/r[ix])
                elif self._x[ix] < 0:
                    theta[ix] = - _np.arcsin(self._z[iz]/r[ix]) + _np.pi

            self.r[:, iz] = (1 - 0.5*(self._cfg.gamma - 1)*Umax**2 \
                             * _np.exp(1-r**2/b**2))**(1/(self._cfg.gamma - 1))
            self.ru[:, iz] -= Umax*r*_np.exp(0.5*(1-r**2/b**2))/b*_np.sin(theta)
            self.rv[:, iz] += Umax*r*_np.exp(0.5*(1-r**2/b**2))/b*_np.cos(theta)

        self.p = self.r**self._cfg.gamma/self._cfg.gamma
        self.ru *= self.r
        self.rv *= self.r

    def poiseuille(self):
        """ Poiseuille """

        # Poiseuille
        tmp_z = _np.arange(self._nz)*self._msh.dz
        self._cfg.U0 *= (4*tmp_z/(tmp_z[-1]-tmp_z[0]) -
                         4*tmp_z**2/(tmp_z[-1]-tmp_z[0])**2)

        for ix in range(self._nx):
            self.ru[ix, :] = self._cfg.U0

    def kelvin_helmholtz(self):
        """
            Initialize fields with a mixing layer with roll-up vortices induced
            by the Kelvin–Helmholtz instability.
        """

        U1 = 0.8
        U2 = 0.2
        delta = 0.4
        T1 = 1.
        T2 = 0.8

        T = _np.zeros((self._nx, self._nz))

        for iz in range(self._nz):
            self.ru[:, iz] = 0.5*((U1+U2) +
                                  (U1-U2)*_np.tanh(2*self._z[iz]/delta))
            T[:, iz] = T1*(self.ru[:, iz] - U2)/(U1 - U2) \
                     + T2*(U1-self.ru[:, iz])/(U1 - U2) \
                     + 0.5*(self._cfg.gamma-1)*(U1 - self.ru[:, iz])*(self.ru[:, iz]-U2)
            self.r[:, iz] = 1/T[:, iz]

        self.p = _np.ones((self._nx, self._nz))/self._cfg.gamma
        self.ru = self.ru*self.r

    def zero_obstacles(self):
        """ Set velocity to 0 in obstacles. """

        for sub in self._msh.obstacles:
            self.ru[sub.sx, sub.sz] = 0
            self.rv[sub.sx, sub.sz] = 0

    def get(self):
        """ Get initial fields as a tuple. """
        return self.p, self.r, self.ru, self.rv, self.re
